# -*- coding: utf-8 -*-
"""
Created on Tue Jul  3 12:10:42 2018

@author: keelin
"""
import numpy as np
import opencxr
from skimage.transform import resize


def rescale_to_min_max(
    np_array_in,
    new_dtype=None,
    new_min=None,
    new_max=None,
):
    """
    A method to rescale array values (image intensities) to a new min and max range.
    if new_dtype is not provided the original dtype will be assumed
    if new_min and/or new_max are not provided then the min/max possible
    for the dtype will be used.
    :param np_array_in: array to rescale
    :param new_dtype: new dtype if required
    :param new_min: new min if required
    :param new_max: new max if required
    :return: The rescaled np array
    """

    int_types = [
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
    ]
    float_types = [np.float32, np.float64]

    if new_dtype is None:
        new_dtype = np_array_in.dtype

    if new_dtype in int_types:
        min_poss = np.iinfo(new_dtype).min
        max_poss = np.iinfo(new_dtype).max
    elif new_dtype in float_types:
        min_poss = np.finfo(new_dtype).min
        max_poss = np.finfo(new_dtype).max
    else:
        raise Exception("Type unsupported for rescaling {}".format(new_dtype))

    if new_min is None:
        new_min = min_poss
    if new_max is None:
        new_max = max_poss

    if new_min < min_poss or new_max > max_poss:
        raise Exception(
            "Arrays of type {} cannot support \
                         min-max {}-{}".format(
                new_dtype, new_min, new_max
            )
        )

    orig_min = np.min(np_array_in)
    orig_max = np.max(np_array_in)
    std = (np_array_in - orig_min) / (orig_max - orig_min)
    scaled = std * (new_max - new_min) + new_min
    return scaled.astype(new_dtype)


def resize_isotropic(
    np_array_img,
    old_spacing,
    new_size_for_axis,
    axis_specified=1,
    anti_aliasing=True,
    interp_order=1,
):
    """
    A method which reads in a 2D np array representing an image and resizes it
    to have isotropic spacing based on the original spacing and the new size
    along one of the dimensions.
    :param np_array_img: Input array
    :param old_spacing: Original image spacing
    :param new_size_for_axis: The new image size specified for a particular axis
    :param axis_specified: The axis for which size is specified (0 or 1)
    :param anti_aliasing: Whether to use anti_aliasing in the resize operation
    :param interp_order: as per skimage transform 0: Nearest-neighbor
        1: Bi-linear (default)
        2: Bi-quadratic
        3: Bi-cubic
        4: Bi-quartic
        5: Bi-quintic
    :return: The resized array
             The new spacing
              The list of size changes carried out for reference or future use (see utils __init__.py)
    """

    orig_dtype = np_array_img.dtype
    new_spacing = (
        old_spacing[axis_specified] * np_array_img.shape[axis_specified]
    ) / new_size_for_axis
    other_axis = 0 if axis_specified == 1 else 1
    new_size_other_axis = round(
        np_array_img.shape[other_axis] * old_spacing[other_axis] / new_spacing
    )
    resized_array = None

    if axis_specified == 0:
        resized_array, returned_spacing, size_changes = resize_to_x_y(
            np_array_img,
            old_spacing,
            new_size_for_axis,
            new_size_other_axis,
            anti_aliasing=anti_aliasing,
            order=interp_order,
        )
    else:
        resized_array, returned_spacing, size_changes = resize_to_x_y(
            np_array_img,
            old_spacing,
            new_size_other_axis,
            new_size_for_axis,
            anti_aliasing=anti_aliasing,
            order=interp_order,
        )

    resized_array = resized_array.astype(orig_dtype)
    return np.squeeze(resized_array), returned_spacing, size_changes


def resize_to_x_y(
    np_array_img,
    old_spacing,
    new_size_0,
    new_size_1,
    anti_aliasing=True,
    interp_order=1,
):
    """
     A method which reads in a 2D np array representing an image and
    resizes to new x y dimensions.
    :param np_array_img: Input image
    :param old_spacing: original image spacing
    :param new_size_0: new size for axis 0
    :param new_size_1: new size for axis 1
    :param anti_aliasing:  Whether to use anti_aliasing in the resize operation
    :param interp_order: as per skimage transform 0: Nearest-neighbor
        1: Bi-linear (default)
        2: Bi-quadratic
        3: Bi-cubic
        4: Bi-quartic
        5: Bi-quintic
    :return: The new resized image
             The new image spacing
             The list of size changes carried out for reference or future use (see utils __init__.py)

    """

    new_img = resize(
        image=np_array_img,
        output_shape=(new_size_0, new_size_1),
        preserve_range=True,
        mode="reflect",
        anti_aliasing=anti_aliasing,
        order=interp_order,
    )
    new_spacing_x = old_spacing[0] * np_array_img.shape[0] / new_size_0
    new_spacing_y = old_spacing[1] * np_array_img.shape[1] / new_size_1

    size_changes = [
        [
            opencxr.utils.size_change_resize_to_x_y,
            [np_array_img.shape[0], np_array_img.shape[1], new_size_0, new_size_1],
        ]
    ]

    return new_img, [new_spacing_x, new_spacing_y], size_changes


def resize_preserve_aspect_ratio(
    np_array_img,
    old_spacing,
    new_size_for_axis,
    axis_specified,
    anti_aliasing=True,
    interp_order=1,
):
    """
     A method to resize an image preserving the aspect ratio.  One new edge size is specified and the axis
    that is to be resized to that size.  The other dimension will be resized accordingly.
    :param np_array_img: The input image
    :param old_spacing: The original image spacing
    :param new_size_for_axis: The length being specified for one axis
    :param axis_specified: The axis to be resized to the given length
    :param anti_aliasing: Whether to use anti-aliasing
    :param interp_order:  as per skimage transform 0: Nearest-neighbor
        1: Bi-linear (default)
        2: Bi-quadratic
        3: Bi-cubic
        4: Bi-quartic
        5: Bi-quintic
    :return: The resized image
             The new image spacing
             The list of size changes carried out for reference or future use (see utils __init__.py)
    """

    if axis_specified == 0:
        mult_factor = new_size_for_axis / np_array_img.shape[0]
        new_shape_0 = new_size_for_axis
        new_shape_1 = int(np.round(np_array_img.shape[1] * mult_factor))
    else:
        mult_factor = new_size_for_axis / np_array_img.shape[1]
        new_shape_0 = int(np.round(np_array_img.shape[0] * mult_factor))
        new_shape_1 = new_size_for_axis

    img_out, new_spacing, size_changes = resize_to_x_y(
        np_array_img, old_spacing, new_shape_0, new_shape_1, anti_aliasing, interp_order
    )

    return img_out, new_spacing, size_changes


def resize_long_edge_and_pad_to_square(
    np_array_img,
    old_spacing,
    square_edge_size,
    pad_value=0,
    anti_aliasing=True,
    interp_order=1,
):
    """
     A method to resize an image to a square size (square_edge_size * square_edge_size) without changing aspect ratio
    The image is first resized, preserving aspect ratio, such that the longer edge is of length square_edge_size
    The shorter edge is then padded (split across left/right or top/bottom) to bring it to the same length.
    :param np_array_img: the input image
    :param old_spacing: the input image spacing
    :param square_edge_size: the output image required edge length
    :param pad_value: the intensity value for padding pixels
    :param anti_aliasing: whether to use anti-aliasing
    :param interp_order: as per skimage transform 0: Nearest-neighbor
        1: Bi-linear (default)
        2: Bi-quadratic
        3: Bi-cubic
        4: Bi-quartic
        5: Bi-quintic
    :return: The resized image
             The new image spacing
             The list of size changes carried out for reference or future use (see utils __init__.py)
    """

    shape_x, shape_y = np_array_img.shape

    if shape_x >= shape_y:
        # do resize
        img_resized, new_spacing, resize_size_changes = resize_preserve_aspect_ratio(
            np_array_img, old_spacing, square_edge_size, 0, anti_aliasing, interp_order
        )

        # pad smaller dimension (in this case top and bottom)
        diff = square_edge_size - img_resized.shape[1]
        img_padded, pad_size_changes = pad_axis_with_total(
            img_resized, axis=1, total_pad=diff
        )

    else:
        # do resize
        img_resized, new_spacing, resize_size_changes = resize_preserve_aspect_ratio(
            np_array_img, old_spacing, square_edge_size, 1, anti_aliasing, interp_order
        )

        # pad smaller dimension (in this case left and right)
        diff = square_edge_size - img_resized.shape[0]
        img_padded, pad_size_changes = pad_axis_with_total(
            img_resized, axis=0, total_pad=diff
        )

    # append the pad_size changes on to the resize_size_changes before returning
    resize_size_changes.extend(pad_size_changes)

    return img_padded, new_spacing, resize_size_changes


def pad_axis_with_total(img_np, axis, total_pad, pad_value=0):
    """
    Pad an image with rows (if axis==1) or columns (if axis==0)
    The total number of rows/columns will be split between the top/bottom or the left/right of the image
    :param img_np: The input image
    :param axis: The axis to pad  (0 pads left/right, 1 pads top/bottom)
    :param total_pad: The total number of rows/columns to be add.  Will be split in half.
    :param pad_value: The value to give to padding rows/cols
    :return: The padded image
             The list of size changes carried out for reference or future use (see utils __init__.py)
    """

    # pad top and bottom
    if axis == 1:
        top_pad = int(np.round(float(total_pad) / 2))
        bottom_pad = total_pad - top_pad
        img_np = np.pad(
            img_np,
            ((0, 0), (top_pad, bottom_pad)),
            "constant",
            constant_values=pad_value,
        )

    elif axis == 0:
        left_pad = int(np.round(float(total_pad) / 2))
        right_pad = total_pad - left_pad
        img_np = np.pad(
            img_np,
            ((left_pad, right_pad), (0, 0)),
            "constant",
            constant_values=pad_value,
        )

    return img_np, [
        [opencxr.utils.size_change_pad_axis_with_total, [axis, total_pad, pad_value]]
    ]


def un_pad_axis_with_total(img_np, axis, total_pad):
    """
    Reverse a previous padding of an image removing rows (if axis==1) or columns (if axis==0)
    The total number of rows/columns removed will be split between the top/bottom or the left/right of the image
    :param img_np: the input image
    :param axis: the axis to (un)pad (0 pads left/right, 1 pads top/bottom)
    :param total_pad: The total number of rows/columns to be un-padded.  Will be split in half.
    :return: The unpadded image
             The list of size changes carried out for reference or future use (see utils __init__.py)
    """
    # unpad top and bottom
    if axis == 1:
        top_pad = int(np.round(float(total_pad) / 2))
        bottom_pad = total_pad - top_pad
        full_height = img_np.shape[1]
        img_np = img_np[:, top_pad : full_height - bottom_pad]

    # unpad left and right
    elif axis == 0:
        left_pad = int(np.round(float(total_pad) / 2))
        right_pad = total_pad - left_pad
        full_width = img_np.shape[0]
        img_np = img_np[left_pad : full_width - right_pad, :]

    return img_np, [
        [opencxr.utils.size_change_unpad_axis_with_total, [axis, total_pad]]
    ]



# -*- coding: utf-8 -*-
"""
@author: keelin
"""

import numpy as np
from opencxr.utils.mask_crop import crop_img_borders
from opencxr.utils.resize_rescale import resize_preserve_aspect_ratio
from scipy import ndimage


class Normalizer:
    """
    Implements the method of Philipsen et al, IEEE Transactions on Medical Imaging, 2015
    https://ieeexplore.ieee.org/document/7073580
    "Localized Energy-Based Normalization of Medical Images: Application to Chest Radiography"

    In general users of opencxr need only access method do_full_normalization()

    The coefficients used here (coeffs70 and coeffs_lungseg) are generated by opencxr team,
    using a private set of 50 CXR images as described in the Philipsen paper
    """

    @classmethod
    def split_energy_bands(cls, image, sigmas):
        """
        Split an np image to energy bands as described in section III.A of Philipsen et al, also reference [40] of that work.
        :param image: the input image
        :param sigmas: The array of sigmas, set as per Philipsen at [1, 2, 4, 8, 16]
        :return: The array of energy band images from 16 to 1, with the filtered image in the last place
        """

        bands = np.zeros((len(sigmas) + 1,) + image.shape)

        curr_I_x = image
        for sigma_ind, sigma in enumerate(sigmas):
            curr_L_x = ndimage.filters.gaussian_filter(curr_I_x, sigma, mode="wrap")

            bands[sigma_ind] = curr_I_x - curr_L_x

            # when we have no sigmas left, store the final filtered image in the last band (Fig 3 and reference [40])
            if sigma_ind == len(sigmas) - 1:
                bands[sigma_ind + 1] = curr_L_x

            curr_I_x = curr_L_x

        # return the bands in reverse order of sigmas ([16,8,4,2,1])
        return bands[::-1]

    @classmethod
    def report_energy_bands(cls, bands, mask=1):
        """
        given the energy bands of the image, and a region to focus on
        returns the means, stdevs of the energy bands in that region
        the stdevs are the "energy values" e_i_omega as used in formula 5 of Philipsen paper
        :param bands: energy bands from split_energy_bands method
        :param mask: region of interest (either central 70% of image, or a lung mask)
        :return: a list of means and a list of std-devs
        """
        means = []
        stdevs = []
        shape = bands[0].shape
        if isinstance(
            mask, np.ndarray
        ):  # if the mask provided is an array (a lung mask)
            for band in bands:
                values = band[mask > 0]
                means.append(values.mean())
                stdevs.append(values.std())
        else:  # otherwise we have no lung mask, just use central 70% of the image
            for band in bands:
                sub_im = band[
                    int(0.15 * shape[0]) : int(0.85 * shape[0]),
                    int(0.15 * shape[1]) : int(0.85 * shape[1]),
                ]
                means.append(sub_im.mean())
                stdevs.append(sub_im.std())
        return means, stdevs

    @classmethod
    def reconstruct(cls, bands, means, stdevs, coefficients):
        """
        reconstructs the normalized image from the energybands and reference coefficients
        see formula 5 of Philipsen et al
        :param bands: the bands from split_energy_bands method
        :param means: the means from report_energy_bands method
        :param stdevs: the stddevs from report_energy_bands method
        :param coefficients: the coefficients which are constants derived from a reference dataset
        :return: The reconstructed normalized image
        """
        bands[0] = (bands[0] - means[0]) / stdevs[0]
        for j in range(1, bands.shape[0]):
            bands[j] = bands[j] * coefficients[j] / stdevs[j]
        return bands.sum(0)

    @classmethod
    def get_norm_central_70(cls, img_np, spacing):
        """
        Get a normalized image based on central 70% of the image
        Will first crop the image (removing homogeneous black border regions) and resize it to 2048 wide (preserving aspect ratio)
        :param img_np: The input image to be normalized
        :param spacing: the spacing of the input image
        :return: A normalized image for feeding to step 2 of the normalization algorithm (using a lung mask)
                 The same normalized image that is rescaled/clipped so that it is nicer to view
                 The spacing of the returned image
                 The list of size changes carried out for reference or future use (see utils __init__.py)
        """
        # the coefficients lambda, calculated by opencxr team, based on a private reference set of 50 images (see formula 4 in Philipsen et al)
        coeffs70 = [1, 0.15046743, 0.09473514, 0.06337214, 0.0451897, 0.03574716]
        # 6 sigma bands as per the Philipsen paper.
        sigmas = sigmas = [1, 2, 4, 8, 16]
        # make sure the image is float
        img_np = img_np.astype(np.float64)
        # crop away black borders
        img_np, size_changes_border_crop = crop_img_borders(
            img_np, in_thresh_factor=0.05
        )
        # resize to width of 2048
        img_np, new_spacing, size_changes_2048 = resize_preserve_aspect_ratio(
            img_np, spacing, 2048, 0
        )
        # split into energy bands
        bands = cls.split_energy_bands(img_np, sigmas)
        # get means and stddevs of the energy bands
        means, stdevs = cls.report_energy_bands(bands)
        # reconstruct an image from the energy bands and the reference values
        norm_70 = cls.reconstruct(bands, means, stdevs, coeffs70)

        # the norm_70 image can be fed to the next normalization step.  But for a nice readable image we can scale/clip it.
        # set to min, max 0,4095
        new_min = 0
        new_max = 4095
        img_mean = norm_70.mean()
        set_min = img_mean - 5.0
        set_max = img_mean + 5.0
        readable_img = np.clip(
            (new_max - new_min) * ((norm_70 - set_min) / (set_max - set_min)) + new_min,
            new_min,
            new_max,
        ).astype(np.uint16)

        # combine all size changes in a single list to return
        size_changes_border_crop.extend(size_changes_2048)

        # return the norm_70 for step2, the readable norm70, the new spacing, and the list of size changes
        return norm_70, readable_img, new_spacing, size_changes_border_crop

    @classmethod
    def get_norm_lung_mask(cls, img_np_norm70, lung_mask_np):
        """
        The second normalization step.  This takes the norm_70 image from get_norm_central_70, and a lung mask of it.
        The second normalization step is applied.
        :param img_np_norm70: The input image, the first image returned by get_norm_central_70()
        :param lung_mask_np: The lung mask of the input image
        :return: the final normalized image, scaled and clipped for best viewing
        """

        # the coefficients lambda based on a reference set of 50 images (see formula 4 in Philipsen et al)
        coeffs_lungseg = [1, 0.26093275, 0.18805708, 0.13976646, 0.1033522, 0.07498657]
        # 6 sigma bands as per the Philipsen paper.
        sigmas = [1, 2, 4, 8, 16]
        # split into energy bands
        bands = cls.split_energy_bands(img_np_norm70, sigmas)
        # get means and stddevs of the energy bands
        means, stdevs = cls.report_energy_bands(bands, mask=lung_mask_np)
        # reconstruct from the energy bands and reference coefficients
        norm = cls.reconstruct(bands, means, stdevs, coeffs_lungseg)

        # rescale and clip to more readable values:
        new_min = 0
        new_max = 4095
        #
        set_min = -5.0
        set_max = 5.0

        # rescale from range -5, +5 to 0,4095
        # and clip values that end up outside that range
        readable_img = np.clip(
            (new_max - new_min) * ((norm - set_min) / (set_max - set_min)) + new_min,
            new_min,
            new_max,
        )

        return readable_img

    @classmethod
    def do_full_normalization(cls, img_in, spacing, lung_seg_algorithm):
        """
        Does full 2 step normalization including an intermediate lung segmentation
        :param img_in: The input image to be normalized
        :param spacing: The spacing of the input image
        :param lung_seg_algorithm: an instance of the lung segmentation algorithm
        :return: The normalized image with larger axis at size 2048 pixels
                  The new spacing for the normalized image
                  The list of size changes carried out for reference or future use (see utils __init__.py)
        """

        # do norm step 1 (central 70%)
        norm_70, readable_img, new_spacing, size_changes = cls.get_norm_central_70(
            img_in, spacing
        )

        # do a lung segmentation on the norm image
        lung_seg_mask = lung_seg_algorithm.run(readable_img)

        # if the lung seg mask is completely empty this is probably not a lung image at all so we should fail gracefully
        # i.e. return an empty image
        if np.max(lung_seg_mask) == 0:
            print("lung seg finds no lung so cxr standardization returning empty image")
            return lung_seg_mask, new_spacing, size_changes

        # do norm step 2 using the lung segmentation image
        final_norm_img = cls.get_norm_lung_mask(norm_70, lung_seg_mask)

        return final_norm_img, new_spacing, size_changes



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 21 11:28:21 2019

@author: keelin
"""

from numpy import fliplr, flipud
from opencxr.utils.resize_rescale import rescale_to_min_max
from skimage import util
from skimage.transform import rotate


def invert_grayscale(np_array_in, preserve_dtype=True):
    """
    A method to invert pixel grayvalues
    If preserve_dtype is true then the result will be returned as the original
    data type.
    Resulting grey-values will be rescaled to min-max values of that dtype
    :param np_array_in: input image
    :param preserve_dtype: whether to preserve the input dtype
    :return: the image with intensities inverted
    """

    inverted_np = util.invert(np_array_in)
    if preserve_dtype:
        # cast back to original type, first rescaling to min/max for that type
        inverted_np = rescale_to_min_max(
            inverted_np,
            np_array_in.dtype,
            new_min=None,
            new_max=None,
        )
    return inverted_np


def rotate_img(np_array_in, rot_angle, preserve_dtype=True):
    """
    A method to rotate clockwise (rot_angle in degrees)
    If preserve_dtype is true then the result will be returned as the original
    data type.
    Resulting grey-values will be rescaled to min-max values of that dtype
    :param np_array_in: input image
    :param rot_angle: angle of rotation
    :param preserve_dtype: whether to preserve input dtype
    :return: The rotated image
    """

    rot_img = rotate(np_array_in, rot_angle)
    if preserve_dtype:
        # cast back to original type, first rescaling to min/max for that type
        rot_img = rescale_to_min_max(
            rot_img, np_array_in.dtype, new_min=None, new_max=None
        )
    return rot_img


def flip_x(np_array_in, preserve_dtype=True):
    """
    A method to flip horizontally (switch image left and right sides)
    If preserve_dtype is true then the result will be returned as the original
    data type.
    Resulting grey-values will be rescaled to min-max values of that dtype
    :param np_array_in: input image
    :param preserve_dtype: whether to preserve dtype
    :return: the horizontally flipped image
    """
    flipx = flipud(np_array_in)
    if preserve_dtype:
        # cast back to original type, first rescaling to min/max for that type
        flipx = rescale_to_min_max(flipx, np_array_in.dtype, new_min=None, new_max=None)
    return flipx


def flip_y(np_array_in, preserve_dtype=True):
    """
    A method to flip vertically (switch image top and bottom)
    If preserve_dtype is true then the result will be returned as the original
    data type.
    Resulting grey-values will be rescaled to min-max values of that dtype
    :param np_array_in: input image
    :param preserve_dtype: whether to preserve dtype
    :return: the vertically flipped image
    """
    flipy = fliplr(np_array_in)
    if preserve_dtype:
        # cast back to original type, first rescaling to min/max for that type
        flipy = rescale_to_min_max(flipy, np_array_in.dtype, new_min=None, new_max=None)
    return flipy



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 11:19:57 2020

@author: keelin
"""

from pathlib import Path

import numpy as np
from keras.models import load_model
from opencxr.algorithms.base_algorithm import BaseAlgorithm
from opencxr.algorithms.imagesorter.preprocess import preprocess_img
import wget
import os

"""
An algorithm to sort 2D images into the following categories:
 - CXR (PA)
 - CXR (AP)
 - CXR (lateral)
 - Non-CXR
In addition, the following information is provided for each image 
 - Rotation (in degrees, based on expectation of an upright CXR image)
 - Lateral flip (for a lateral CXR image, based on expectation of spine to right of lateral image)
 - Inversion (based on expectation of CXR image with dark intensitities for air filled regions)
"""


class ImageSorterAlgorithm(BaseAlgorithm):
    def name(self):
        return "ImageSorterAlgorithm"

    def __init__(self):
        """
        load the model
        """
        path_to_model_file = (
            Path(__file__).parent.parent / "model_weights" / "image_sorter.hdf5"
        )
        path_to_model_resolved = str(path_to_model_file.resolve())

        file_url = "https://github.com/DIAGNijmegen/opencxr/raw/master/opencxr/algorithms/model_weights/image_sorter.hdf5"
        # if the file does not exist (it's not included in whl file) then download it from github
        if not os.path.isfile(path_to_model_resolved):
            print("First use of imagesorter model, downloading the weights......")
            os.makedirs(os.path.dirname(path_to_model_resolved), exist_ok=True)
            wget.download(file_url, path_to_model_resolved)
            if not os.path.isfile(path_to_model_resolved):
                print("Failed to download file from", file_url)
                print(
                    "Please check the URL is valid and the following location is writeable",
                    path_to_model_resolved,
                )
                return
        if (
            not os.stat(path_to_model_resolved).st_size > 150000000
        ):  # model file needs to be larger than 150MB or it was not pulled correctly
            print(
                "ImageSorter model file seems incorrect (size ",
                os.stat(path_to_model_resolved).st_size,
                ") downloading the weights......",
                path_to_model_resolved,
            )
            os.remove(
                path_to_model_resolved
            )  # first delete the file, because wget will not overwrite
            wget.download(file_url, path_to_model_resolved)

        self.model = load_model(path_to_model_resolved)

    def run(self, image):
        """
        Run the image sorter algorithm
        :param image: The input image in x,y axis order (as from utils file_io read_image())
        :return: # a dict something like:
        {'Type': 'PA',
           'Rotation': '0',
           'Inversion': 'No',
           'Lateral_Flip': 'No',
           'Type_Probs_PA_AP_lateral_notCXR': [0.99999976, 2.5101654e-08, 2.4382584e-07, 1.0590604e-08],
           'Rotation_Probs_0_90_180_270': [0.9999999, 2.7740466e-08, 2.2800064e-08, 3.7591672e-08],
           'Inversion_Probs_No_Yes': [0.9999968589511354, 3.1410489e-06],
           'Lateral_Flip_Probs_No_Yes': [0.9999986753330177, 1.324667e-06]}
        The first four keys give classifications for Type, Rotation, Inversion, Lateral_Flip.
        The second four keys provide probabilities of all possible classes for users that might need this

        Possible values for the first four keys are as follows:
        Type:  ['PA', 'AP', 'lateral', 'notCXR']
        Rotation: ['0', '90', '180', '270']
        Inversion: ['No', 'Yes']
        Lateral_Flip: ['No', 'Yes']
        """
        # first preprocessing:
        image = preprocess_img(image)

        # now get ready and call the model to do prediction
        image = image / 255.0
        image = np.stack((image, image, image), axis=2)
        pred = self.model.predict(np.expand_dims(image, 0))

        # binarize the model outputs
        im_type = np.argmax(pred[0])
        im_rot = np.argmax(pred[1])
        im_inv = 1 if pred[2] > 0.5 else 0
        im_flip = 1 if pred[3] > 0.5 else 0

        # set up return labels
        type_labels = ["PA", "AP", "lateral", "notCXR"]
        rotation_labels = ["0", "90", "180", "270"]
        inversion_labels = ["No", "Yes"]
        lateral_flip_labels = ["No", "Yes"]

        # return dict
        return {
            "Type": type_labels[im_type],
            "Rotation": rotation_labels[im_rot],
            "Inversion": inversion_labels[im_inv],
            "Lateral_Flip": lateral_flip_labels[im_flip],
            "Type_Probs_PA_AP_lateral_notCXR": list(pred[0][0]),
            "Rotation_Probs_0_90_180_270": list(pred[1][0]),
            "Inversion_Probs_No_Yes": [1 - pred[2][0][0], pred[2][0][0]],
            "Lateral_Flip_Probs_No_Yes": [1 - pred[3][0][0], pred[3][0][0]],
        }



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 21 12:11:36 2019

@author: keelin
"""
import numpy as np
import opencxr.utils
import skimage.feature
from opencxr.utils.resize_rescale import rescale_to_min_max
from scipy import ndimage
from scipy.ndimage import binary_dilation, find_objects
from skimage import measure

"""
Methods relating to mask and crop functionality
"""


def set_non_mask_constant(img_np, mask_np, dilation_in_pixels=0, constant_val=0):
    """
    A method to set a portion of an image to a constant value
    e.g. set areas outside lungs to be black
    Allows for dilation of the mask if required
    :param img_np: the original image
    :param mask_np: the mask, e.g. lung mask, expected as binary 1 or 0 values
    :param dilation_in_pixels: the size of the dilation of the mask
    :param constant_val: the value to set non-mask as
    :return: The image as np with non mask areas set to the constant value
    """
    if dilation_in_pixels > 0:
        struct_element = np.ones((dilation_in_pixels, dilation_in_pixels)).astype(bool)
        mask_np = binary_dilation(mask_np, structure=struct_element).astype(
            mask_np.dtype
        )

    # make a copy because putmask will change the original
    img_np_copy = np.copy(img_np)
    np.putmask(img_np_copy, mask_np < 1, constant_val)
    return img_np_copy


def crop_to_mask(img_np, spacing, mask_np, margin_in_mm):
    """
    A method to crop away regions outside the bounding box of a mask
    e.g. crop to smallest rectangle containing lungs
    :param img_np: the original image
    :param spacing: The spacing of the original image
    :param mask_np: the mask to crop around, e.g. lung mask,
               expected as binary 1,0 values
    :param margin_in_mm: a margin to allow around the tightest bounding box
    :return: The cropped image
             The size changes list for reference or future use see utils __init__.py
    """
    # convert margin in mm to margin_in_pixels for each of x and y
    margin_in_pixels_x = int(np.round(margin_in_mm / spacing[0]))
    margin_in_pixels_y = int(np.round(margin_in_mm / spacing[1]))

    # get bounding box for mask
    bbox = find_objects(mask_np)
    min_x_mask = max(bbox[0][0].start - margin_in_pixels_x, 0)
    min_y_mask = max(bbox[0][1].start - margin_in_pixels_y, 0)
    max_x_mask = min(bbox[0][0].stop + margin_in_pixels_x, mask_np.shape[0])
    max_y_mask = min(bbox[0][1].stop + margin_in_pixels_y, mask_np.shape[1])

    cropped_img, size_changes = crop_with_params(
        img_np, [min_x_mask, max_x_mask, min_y_mask, max_y_mask]
    )
    return cropped_img, size_changes


def crop_img_borders_by_edginess(
    img_np_array, width_edgy_threshold=50, dist_edgy_threshold=100
):
    """
    Method to crop homogeneous border regions based on edge detection:
    1) do an edge detection to pick up edges belonging to the image content
    2) crop to the region which contains edge information
                                        (exclude small isolated edgy regions)
    :param img_np_array: the input image
    :param width_edgy_threshold: the threshold (in pixels) to define
                           an edgy region as 'small'
    :param dist_edgy_threshold: the threshold (in pixels) to define an edgy region as
                          'distant' from other edgy regions
    :return: The cropped image
             The size changes list for reference or future use see utils __init__.py
    """

    def find_starts_ends_edgy_regions_axis(edge_img_np_array, axis_to_check):
        """
        inner function to count edge pixels and identify (per row, per axis)
        where the regions with edges start and end
        """
        count_edges = []
        start_end_edgy_regions = []
        for ind in range(0, edge_img_np_array.shape[axis_to_check]):
            count_edge_pixels = np.sum(
                edge_img_np_array[ind, :]
                if axis_to_check == 0
                else edge_img_np_array[:, ind]
            )
            count_edges.append(count_edge_pixels)
            if ind == 0:
                # start an edgy region if there are immediately edges present
                if count_edge_pixels > 0:
                    start_end_edgy_regions.append(ind)
                continue
            if ind == img_np_array.shape[axis_to_check] - 1:
                # if previous one was non-zero we were in an edgy region
                # so add the ending
                if count_edges[ind - 1] > 0:
                    start_end_edgy_regions.append(ind)
                # if the last row is an isolated edgy region
                # add it as a start and end also.
                elif count_edges[ind - 1] == 0 and count_edge_pixels > 0:
                    start_end_edgy_regions.append(ind)
                    start_end_edgy_regions.append(ind)
                continue

            # otherwise we are somewhere in the middle
            if count_edge_pixels > 0 and count_edges[ind - 1] == 0:
                # add start location for edgy region
                start_end_edgy_regions.append(ind)
            elif count_edge_pixels == 0 and count_edges[ind - 1] > 0:
                # add end location for edgy region
                start_end_edgy_regions.append(ind)

        return start_end_edgy_regions

    def remove_small_isolated_edgy_regions(starts_ends_edgy_regions):
        """
        inner function to remove the edgy regions which are small or isolated
        """
        # print('starts and ends are ', starts_ends_edgy_regions)
        starts_ends_retain = []
        for start_edgy_index in range(0, len(starts_ends_edgy_regions), 2):
            start_edgy = starts_ends_edgy_regions[start_edgy_index]
            end_edgy = starts_ends_edgy_regions[start_edgy_index + 1]
            # print("found start and end", start_edgy, end_edgy)
            width_edgy = end_edgy - start_edgy + 1
            # print("found width ", width_edgy)

            dist_next_edgy = 10000
            dist_prev_edgy = 10000
            # if a subsequent edgy region exists
            if start_edgy_index + 2 < len(starts_ends_edgy_regions):
                start_next_edgy = starts_ends_edgy_regions[start_edgy_index + 2]
                dist_next_edgy = start_next_edgy - end_edgy
            # if a previous edgy region exists
            if start_edgy_index - 2 >= 0:
                end_prev_edgy = starts_ends_edgy_regions[start_edgy_index - 1]
                dist_prev_edgy = start_edgy - end_prev_edgy

            isolated_left = dist_prev_edgy > dist_edgy_threshold
            isolated_right = dist_next_edgy > dist_edgy_threshold

            is_small_edgy_region = width_edgy < width_edgy_threshold
            is_isolated_edgy_region = isolated_left and isolated_right

            if not (is_small_edgy_region and is_isolated_edgy_region):
                starts_ends_retain.append(start_edgy)
                starts_ends_retain.append(end_edgy)
        # print('starts and ends i will retain are ', starts_ends_retain)
        return starts_ends_retain

    # Now start the main work:
    # Parameters of Canny are based on the input being in range 0-65535
    # so need to force this before we run Canny
    img_for_edge_det = rescale_to_min_max(img_np_array, new_dtype=np.uint16)

    edge_img = skimage.feature.canny(
        image=img_for_edge_det.astype(np.float32),
        sigma=5.0,
        low_threshold=0.0,
        high_threshold=500.0,
    )
    # convert from boolean
    edge_img = edge_img.astype(np.uint8)

    # Now crop according to where the "edgy" region of the image is
    start_end_edges_x = find_starts_ends_edgy_regions_axis(edge_img, 0)
    start_end_edges_x = remove_small_isolated_edgy_regions(start_end_edges_x)

    start_end_edges_y = find_starts_ends_edgy_regions_axis(edge_img, 1)
    start_end_edges_y = remove_small_isolated_edgy_regions(start_end_edges_y)

    start_x = start_end_edges_x[0]
    end_x = start_end_edges_x[len(start_end_edges_x) - 1]
    start_y = start_end_edges_y[0]
    end_y = start_end_edges_y[len(start_end_edges_y) - 1]

    # print('Will finally crop from edginess x', start_x, end_x)
    # print('Will finally crop from edginess y', start_y, end_y)

    cropped_img, size_changes = crop_with_params(
        img_np_array, [start_x, end_x, start_y, end_y]
    )
    return cropped_img, size_changes


def crop_img_borders(img_np_array, in_thresh_factor=0.05):
    """
    Method to crop homogeneous border regions based on std deviations:
    :param img_np_array: the input image
    :param in_thresh_factor: threshold factor -> 0.05 recommended
    :return: The cropped image
             The size changes list for reference or future use see utils __init__.py
    """
    xmin = 0
    xmax = img_np_array.shape[0]
    ymin = 0
    ymax = img_np_array.shape[1]

    # Use the image std dev and in_thresh_factor to make a hard threshold
    img_std_dev = np.std(img_np_array)
    hard_threshold = in_thresh_factor * img_std_dev

    # Some setup
    completed = False
    xmin_stored = 0
    xmax_stored = img_np_array.shape[0] - 1
    ymin_stored = 0
    ymax_stored = img_np_array.shape[1] - 1

    # Loop until no more changes can be made
    while not completed:
        xmin = xmin_stored
        xmax = xmax_stored
        ymin = ymin_stored
        ymax = ymax_stored

        # Determine xmin
        # the first x value where the column of pixels is not "homogeneous"
        for x_pix in range(xmin_stored, xmax_stored + 1):
            line = img_np_array[x_pix, ymin_stored : ymax_stored + 1]
            std = np.std(line)
            if std < hard_threshold:
                xmin = x_pix
            else:
                break

        # Determine xmax
        # the last x value where the column of pixels is not "homogeneous"
        for x_pix in range(xmax_stored, -1, -1):
            line = img_np_array[x_pix, ymin_stored : ymax_stored + 1]
            std = np.std(line)
            if std < hard_threshold:
                xmax = x_pix
            else:
                break

        # Determine ymin
        # the first y value where the row of pixels is not "homogeneous"
        for y_pix in range(ymin_stored, ymax_stored + 1):
            line = img_np_array[xmin_stored : xmax_stored + 1, y_pix]
            std = np.std(line)
            if std < hard_threshold:
                ymin = y_pix
            else:
                break

        # Determine ymax
        # the last x value where the row of pixels is not "homogeneous"
        for y_pix in range(ymax_stored, -1, -1):
            line = img_np_array[xmin_stored : xmax_stored + 1, y_pix]
            std = np.std(line)
            if std < hard_threshold:
                ymax = y_pix
            else:
                break

        # In case of any unusual image causing issues - just keep stored values
        if xmax < xmin:
            xmin = xmin_stored
        if ymax < ymin:
            ymin = ymin_stored

        # Assume that we are done, but if we made any change on this
        # last loop then we will set completed=False and try one more time
        completed = True
        if xmin > xmin_stored:
            xmin_stored = xmin
            completed = False
        if xmax < xmax_stored:
            xmax_stored = xmax
            completed = False
        if ymin > ymin_stored:
            ymin_stored = ymin
            completed = False
        if ymax < ymax_stored:
            ymax_stored = ymax
            completed = False

    cropped_img, size_changes = crop_with_params(
        img_np_array, [xmin_stored, xmax_stored + 1, ymin_stored, ymax_stored]
    )
    return cropped_img, size_changes


def crop_with_params(img_np, array_minx_maxx_miny_maxy):
    """
    Crops an image accoring to the 4 params provided in the array
    :param img_np: the image to be cropped
    :param array_minx_maxx_miny_maxy: the array of crop indices min_x, max_x, min_y, max_y
    :return: The cropped image
             The size changes list for reference or future use see utils __init__.py

    """
    minx = array_minx_maxx_miny_maxy[0]
    maxx = array_minx_maxx_miny_maxy[1]
    miny = array_minx_maxx_miny_maxy[2]
    maxy = array_minx_maxx_miny_maxy[3]
    size_changes = [
        [
            opencxr.utils.size_change_crop_with_params,
            [img_np.shape[0], img_np.shape[1], minx, maxx, miny, maxy],
        ]
    ]
    # print('in crop_with_params, orig size', img_np.shape)
    cropped_img = img_np[minx:maxx, miny:maxy]
    # print('in crop_with_params, cropped size is ', cropped_img.shape)
    return cropped_img, size_changes


def uncrop_with_params(
    img_np, orig_size_x, orig_size_y, array_minx_maxx_miny_maxy, pad_value=0
):
    """
    Undo a previously applied cropping operation
    :param img_np: the input image
    :param orig_size_x: the original x size (we will return to this)
    :param orig_size_y: the original y size (we will return to this)
    :param array_minx_maxx_miny_maxy: the array of crop parameters that was used
    :param pad_value: the value to pad with when we make the image larger (uncrop)
    :return: The uncropped (padded) image
             The size changes list for reference or future use see utils __init__.py
    """

    # the image being passed in was already cropped with the array info given, need to restore it
    minx = array_minx_maxx_miny_maxy[0]
    maxx = array_minx_maxx_miny_maxy[1]
    miny = array_minx_maxx_miny_maxy[2]
    maxy = array_minx_maxx_miny_maxy[3]
    pad_top = miny
    pad_bottom = orig_size_y - maxy
    pad_left = minx
    pad_right = orig_size_x - maxx

    # pad top and bottom
    img_np = np.pad(
        img_np,
        ((pad_left, pad_right), (pad_top, pad_bottom)),
        "constant",
        constant_values=pad_value,
    )

    # if not img_np.shape[0] == orig_size_x:
    #    print('x mismatch', img_np.shape[0], orig_size_x)
    # if not img_np.shape[1] == orig_size_y:
    #    print('y mismatch', img_np.shape[1], orig_size_y)

    size_changes = [
        [
            opencxr.utils.size_change_uncrop_with_params,
            [orig_size_x, orig_size_y, minx, maxx, miny, maxy],
        ]
    ]
    return img_np, size_changes


def get_largest_components(np_array, nr_components):
    """
    Get largest connected components (how many defined by nr_components) from a binary image, background==0
    :param np_array: a binary np array (e.g. segmentation mask)
    :param nr_components: the number of largest components to keep
    :return: a binary np array with only the largest components remaining
    """
    # get a label per connected component, background is 0
    labels = measure.label(np_array, background=0)
    # count pixels per label
    unique, counts = np.unique(labels, return_counts=True)

    # helper function
    def get_key(item):
        return item[1]

    # get ordered list of labels
    counts = sorted(zip(unique, counts), key=get_key, reverse=True)
    # get largest of labels excluding 0, and keeping only the top nr_components items
    largest_labels = [c[0] for c in counts if c[0] != 0][0:nr_components]

    if (
        len(largest_labels) == nr_components
    ):  # if we got the correct number of components out
        out_array = np.full(np_array.shape, False, dtype=bool)
        all_labels_np = np.asarray(labels)
        for l in largest_labels:
            out_array = out_array | (all_labels_np == l)
        out_array = out_array.astype(np.uint8)
        # print('returning correct number of components')
        return out_array
    else:  # otherwise there were not even enough components to reach nr_components so just return the original
        # print('returning fewer than expected components')
        return np_array


def tidy_segmentation_mask(seg_mask, nr_components_to_keep):
    """
    Fill holes and then keep only the largest components (e.g. 2 components for frontal lung seg, 1 component for heart seg)
    :param seg_mask: The segmentation mask to be tidied
    :param nr_components_to_keep: The number of components to keep in the final segmentation mask
    :return: The final tidied segmentation mask
    """
    # make binary
    seg_mask = rescale_to_min_max(seg_mask, np.uint8, 0, 1)

    # fill holes
    seg_mask = ndimage.binary_fill_holes(seg_mask)

    # Only keep the largest components
    seg_mask = get_largest_components(seg_mask, nr_components_to_keep).astype(np.uint8)

    seg_mask = rescale_to_min_max(seg_mask, np.uint8, 0, 255)
    return seg_mask



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 12:38:05 2020

@author: keelin
"""


class BaseAlgorithm:
    """
    In case we want common functionality for all algorithms
    """

    def __init__(self):
        pass  # stuff to run on instance creation goes here



# -*- coding: utf-8 -*-
"""

@author: ecem
"""

from pathlib import Path
import numpy as np
import opencxr
from opencxr.utils.file_io import read_file, write_file


def test_lung_seg():
    # Load the algorithm
    lungseg_algorithm = opencxr.load(opencxr.algorithms.lung_seg)
    # read an image from disk
    f_in = Path(__file__).parent / "resources" / "images" / "c0005.mha"
    f_in = str(f_in.resolve())
    img_np, spacing, pydata = read_file(f_in)
    # run the lung segmentation algorithm on the image (note: in the wild it may be best to perform cxr standardization first)
    # this will return the segmentation map image
    seg_map = lungseg_algorithm.run(img_np)
    # write the output segmentation to disk
    f_out = (
        Path(__file__).parent / "resources" / "tmp_test_outputs" / "c0005_lungseg.mha"
    )
    f_out = str(f_out.resolve())
    write_file(f_out, seg_map, spacing)

    passed = seg_map.shape == img_np.shape and not np.max(seg_map) == 0

    if passed:
        print("Lung Segmentation test completed successfully")
    else:
        print("Lung Segmentation results not as expected")

    assert passed


if __name__ == "__main__":
    test_lung_seg()



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 13:31:04 2020

@author: keelin
"""

import importlib
from opencxr.algorithms.base_algorithm import BaseAlgorithm

# names to identify various algorithms available
image_sorter = "opencxr.algorithms.imagesorter"
lung_seg = "opencxr.algorithms.lungsegmentation"
heart_seg = "opencxr.algorithms.heartsegmentation"
cxr_standardize = "opencxr.algorithms.cxrstandardization"


def create_algorithm(algorithm_modulename):
    """
    Create an instance of a specified algorithm
    :param algorithm_modulename: the algorithm name (from specified list in this file)
    :return: an instance of the algorithm required
    """
    # import the specified algorithm module
    algorithmlib = importlib.import_module(algorithm_modulename)

    # In the file, the class called algorithmfilenamealgorithm() will
    # be instantiated. It has to be a subclass of BaseAlgorithm,
    # and it is case-insensitive.
    algorithm = None
    # take the last part of the algorithm name, e.g. 'lungsegmentation'
    algorithm_name = algorithm_modulename.split(".")[2]
    # add the term 'algorithm' to the end of it
    target_algorithm_name = algorithm_name + "algorithm"
    # look for the appropriate algorithm class in the imported module
    for name, cls in algorithmlib.__dict__.items():
        if name.lower() == target_algorithm_name.lower() and issubclass(
            cls, BaseAlgorithm
        ):
            algorithm = cls

    if algorithm is None:
        print(
            "In %s.py, there should be a subclass of BaseAlgorithm with class name that matches %s in lowercase."
            % (algorithm_filename, target_algorithm_name)
        )
        exit(0)

    # get an instance of the identified algorithm
    instance = algorithm()

    print("algorithm [%s] was created" % (instance.name()))
    return instance



# -*- coding: utf-8 -*-
"""

@author: keelin
"""
import os.path
from pathlib import Path
from opencxr.utils.file_io import read_file, write_file
from opencxr.utils.mask_crop import (
    crop_img_borders,
    crop_img_borders_by_edginess,
    uncrop_with_params,
    set_non_mask_constant,
)
from opencxr.utils.rotate_flip_invert import (
    rotate_img,
    flip_x,
    flip_y,
    invert_grayscale,
)
from opencxr.utils import reverse_size_changes_to_img
import numpy as np


def test_file_io():
    # read files of type dicom, png (8 and 16 bit), mha
    f_in_dcm = Path(__file__).parent / "resources" / "images" / "c0004.dcm"
    f_in_dcm = str(f_in_dcm.resolve())
    img_np_dcm, spacing_dcm, tags = read_file(f_in_dcm)

    f_in_png_8 = Path(__file__).parent / "resources" / "images" / "c0001_8bit.png"
    f_in_png_8 = str(f_in_png_8.resolve())
    img_np_png_8, _, _ = read_file(f_in_png_8)

    f_in_png_16 = Path(__file__).parent / "resources" / "images" / "c0002_16bit.png"
    f_in_png_16 = str(f_in_png_16.resolve())
    img_np_png_16, _, _ = read_file(f_in_png_16)

    f_in_mha = Path(__file__).parent / "resources" / "images" / "c0003.mha"
    f_in_mha = str(f_in_mha.resolve())
    img_np_mha, spacing_mha, _ = read_file(f_in_mha)

    # verify the appropriate information is not empty
    passing = True
    if not img_np_dcm.shape == (2992, 2991):
        print("dcm wrong shape", img_np_dcm.shape)
        passing = False
    if not np.array_equal(spacing_dcm, np.asarray([0.143, 0.143, 1.0])):
        print("dcm wrong spacing", spacing_dcm)
        passing = False
    if (
        not tags["StudyInstanceUID"].value
        == "9999.324256108382831452380358921215687044879"
    ):
        print("dcm tags wrong study UID", tags["StudyInstanceUID"].value)
        passing = False

    if not img_np_png_8.shape == (1024, 1024):
        print("png 8 bit wrong shape", img_np_png_8.shape)
        passing = False

    if not img_np_png_16.shape == (3000, 2967):
        print("png 16 bit wrong shape", img_np_png_16.shape)
        passing = False

    if not img_np_mha.shape == (2834, 2851):
        print("mha wrong shape", img_np_mha.shape)
        passing = False
    if not np.array_equal(spacing_mha, np.asarray([0.148, 0.148])):
        print("mha wrong spacing", spacing_mha)
        passing = False

    # write files of type mha, png (uint8) and png (uint16)
    f_out = Path(__file__).parent / "resources" / "tmp_test_outputs" / "writing_mha.mha"
    f_out = str(f_out.resolve())
    write_file(f_out, img_np_mha, spacing_mha)
    if not os.path.isfile(f_out):
        print("Failed to write file at ", f_out)
        passing = False

    f_out = (
        Path(__file__).parent
        / "resources"
        / "tmp_test_outputs"
        / "writing_png_8bit.png"
    )
    f_out = str(f_out.resolve())
    write_file(f_out, img_np_png_8, [1.0, 1.0])
    if not os.path.isfile(f_out):
        print("Failed to write file at ", f_out)
        passing = False

    f_out = (
        Path(__file__).parent
        / "resources"
        / "tmp_test_outputs"
        / "writing_png_16bit.png"
    )
    f_out = str(f_out.resolve())
    write_file(f_out, img_np_png_16, [1.0, 1.0])
    if not os.path.isfile(f_out):
        print("Failed to write file at ", f_out)
        passing = False

    assert passing


def test_mask_crop():
    passing = True
    # read in an image with a black border present
    f_in = Path(__file__).parent / "resources" / "images" / "c3574.png"
    f_in = str(f_in.resolve())
    img_np, _, _ = read_file(f_in)

    # test first crop_img method
    cropped_img, size_changes = crop_img_borders(img_np)
    if not cropped_img.shape == (1024, 984):
        print("crop_img_borders returned wrong shape", cropped_img.shape)
        passing = False

    # test second crop_img method
    cropped_img, size_changes = crop_img_borders_by_edginess(img_np)
    if not cropped_img.shape == (1022, 984):
        passing = False

    # test uncrop with params method
    img_np_uncropped, _ = reverse_size_changes_to_img(
        cropped_img, [1.0, 1.0], size_changes
    )

    if not img_np_uncropped.shape == (1024, 1024):
        print("uncropped shape not as expected", img_np_uncropped.shape)
        passing = False

    # test set non-mask as constant
    mask_np = np.zeros(img_np.shape)
    mask_np[500:600, 500:600] = 1

    masked_img = set_non_mask_constant(img_np, mask_np)

    # verify masked image is correct size
    if not masked_img.shape == img_np.shape:
        print("masked image has wrong shape", masked_img.shape)
        passing = False
    # verify pixels we reset are zero now and were not in original
    test_patch_masked = masked_img[0:450, 0:450]
    test_patch_unmasked = img_np[0:450, 0:450]
    if not np.all(test_patch_masked == 0):
        print("mask not applied correctly")
        passing = False
    if np.all(test_patch_unmasked == 0):
        print("orig image was already zero in test patch")
        passing = False

    assert passing


def test_rotate_flip_invert():
    passing = True
    f_in = Path(__file__).parent / "resources" / "images" / "c0005.mha"
    f_in = str(f_in.resolve())
    img_np, spacing, _ = read_file(f_in)

    rot90 = rotate_img(img_np, 90)
    rot180 = rotate_img(img_np, 180)
    rot270 = rotate_img(img_np, 270)
    rot360 = rotate_img(img_np, 360)

    flipyx = flip_y(flip_x(img_np))

    flipxyrot90 = flip_x(flip_y(rot90))

    # compare images we expect to be the same
    diff_img = img_np.astype(float) - rot360.astype(float)
    avg_pixel_diff = np.sum(np.abs(diff_img)) / (img_np.shape[0] * img_np.shape[1])
    if not avg_pixel_diff < 20:
        print("rot360 not the same as orig", avg_pixel_diff)
        passing = False

    diff_img = flipyx.astype(float) - rot180.astype(float)
    avg_pixel_diff = np.sum(np.abs(diff_img)) / (img_np.shape[0] * img_np.shape[1])
    if not avg_pixel_diff < 20:
        print("flipyx not same as rot180", avg_pixel_diff)
        passing = False

    diff_img = flipxyrot90.astype(float) - rot270.astype(float)
    avg_pixel_diff = np.sum(np.abs(diff_img)) / (img_np.shape[0] * img_np.shape[1])
    if not avg_pixel_diff < 20:
        print("flipxyrot90 not same as rot270", avg_pixel_diff)
        passing = False

    assert passing


if __name__ == "__main__":
    test_file_io()
    test_mask_crop()
    test_rotate_flip_invert()



# -*- coding: utf-8 -*-
"""
Created on Tue Sep  3 13:58:39 2019
"""

import os

import SimpleITK as sitk
import numpy as np
import png  # pypng library to allow management of 16 bit png files
import pydicom
from skimage import color

"""
Methods to read and write CXR images of various formats
By convention we return (and expect) CXR images as numpy arrays where
shape[0] is the x-dimension (width) of the image
while shape[1] is they y-dimension (height)
"""


def read_file(in_img_location):
    """
    Wrapper method that checks which type of image it is and
    uses the appropriate method to read it. Supports mha, mhd, png, dicom
    :param in_img_location: The file location to be read
    :return: img_np: The image as an np array where shape[0] is the x-dimension (width) of the image
               while shape[1] is they y-dimension (height)
    """
    in_img_location = in_img_location.strip()
    if not os.path.isfile(in_img_location):
        raise Exception("Unable to find file at {}".format(in_img_location))
    extension = os.path.splitext(in_img_location)[1]
    extension = extension.lower()
    if extension == ".mhd" or extension == ".mha":
        return read_mhd_mha(in_img_location)
    elif extension == ".dcm":
        return read_dicom(in_img_location)
    elif extension == ".png":
        return read_png(in_img_location)
    else:
        raise Exception(
            "Do not recognize extension {} in location {}. \
        Unable to read file".format(
                extension, in_img_location
            )
        )


def read_png(in_img_location):
    """
    Method to read in a png image and return it as an np array
    :param in_img_location: The file location for the png image
    :return: The image as np array
             The spacing of the image
             An empty string as a place filler since no dicom tags available
    """
    # sitk now seems to handle png even better than pypng, which has some issues, so moving to sitk
    # weird images to debug with include chest-xray-14 image 00000317_000.png or /BIMCV-COVID19/sub-S03214/ses-E07979
    itk_img = sitk.ReadImage(in_img_location)
    init_np_array = np.squeeze(sitk.GetArrayFromImage(itk_img))

    # Can only handle input images with 2 or 3 axes
    if len(init_np_array.shape) < 2 or len(init_np_array.shape) > 3:
        print("unable to proceed(1), ", im, " img shape is ", init_np_array.shape)
        return
    # If 3 axes then assume one of them is a channel axis, need to make sure it is in last position
    if len(init_np_array.shape) == 3:
        channel_axis = np.argmin(init_np_array.shape)
        # print('channel axis seems to be at ', channel_axis, init_np_array.shape)
        if channel_axis == 0:  # swap axes as channel expected last
            init_np_array = np.swapaxes(init_np_array, 0, 2)
            init_np_array = np.swapaxes(init_np_array, 0, 1)
            # print('swapped axes, shape now is ', init_np_array.shape)
        # if there are 4 channels then we assume we have rgba image.... make it grayscale
        if init_np_array.shape[2] == 4:
            init_np_array = color.rgb2gray(color.rgba2rgb(init_np_array))
            # print('did 2 transforms, shape now is ', init_np_array.shape)
        # if there are 3 channels then we assume we have rgb image.... make it grayscale
        elif init_np_array.shape[2] == 3:
            init_np_array = color.rgb2gray(init_np_array)
            # print('did 1 transform shape now is ', init_np_array.shape)
        else:
            print("unable to proceed(2), ", im, " img shape is ", init_np_array.shape)
            return

    # return with x, y (width, height) axis ordering not numpy usual ordering
    img_np_x_y = np.transpose(init_np_array)
    spacing_x_y = np.transpose(itk_img.GetSpacing())

    return img_np_x_y, spacing_x_y, ""


def read_mhd_mha(in_img_location):
    """
    A method which reads in an mhd or mha image and returns it in
    X(axis 0), Y(axis 1) format with spacing
    :param in_img_location: The file location of the original image
    :return: The image as np array
             The spacing of the image
             An empty string as a place filler since no dicom tags available
    """
    itk_img = sitk.ReadImage(in_img_location)

    img_np_x_y = np.squeeze(np.transpose(sitk.GetArrayFromImage(itk_img)))
    spacing_x_y = np.transpose(itk_img.GetSpacing())

    return img_np_x_y, spacing_x_y, ""


def read_dicom(in_img_location):
    """
    Reads a dicom image
    :param in_img_location: The file location of the dicom image
    :return: The image as np array (x, y, width, height)
             The spacing of the image
             The dicom tags as provided by pydicom
    """
    """
    Reads dicom (twice).
    First using sitk - to get the pixel data.  This is reliable also if the
    data has been jpeg2000 compressed (where pydicom does not work)
    Secondly, using pydicom, to get the headers.  We explicitly remove the
    pydicom version of PixelData to avoid confusion.
    #TODO: Could use sitk to read dicom tag info?
    return np_img_array, spacing, pydicom data
    """
    itk_image = sitk.ReadImage(in_img_location)
    img_np_x_y = np.squeeze(np.transpose(sitk.GetArrayFromImage(itk_image)))
    spacing_x_y = np.transpose(itk_image.GetSpacing())
    # print('spacing from sitk image is', spacing_x_y)

    # Need this for headers, but pydicom cannot read pixel data if it is
    # stored in jpeg2000 compression
    pydicom_version = pydicom.dcmread(
        in_img_location
    )  # try this??? ds = pydicom.filereader.read_partial(f, stop_when=pixel_data_reached)

    # Make a correction because sitk seems to read pixel spacing only from the ImagerPixelSpacing dicom tag
    # In some images that tag is absent but the tag PixelSpacing is available
    if (
        "PixelSpacing" in pydicom_version
        and not "ImagerPixelSpacing" in pydicom_version
    ):
        # print('PixelSpacing dicom tag is ', pydicom_version.PixelSpacing)
        spacing_x_y = pydicom_version.PixelSpacing

    # Remove the pixel data from the pydicom data so only headers are returned
    del pydicom_version.PixelData
    return img_np_x_y, spacing_x_y, pydicom_version


def write_file(out_img_location, img_np_x_y, spacing_x_y=[1.0, 1.0]):
    """
    Wrapper method that checks file extension and uses the appropriate method to write it
    Note writing dicom is not supported
    :param out_img_location: The file location to write to (folders will be created if needed)
    :param img_np_x_y: The image to write in np array (x, y, width, height)
    :param spacing_x_y: The spacing of the image (if none then 1.0, 1.0 assumed)
    :return:
    """

    out_img_dir = os.path.dirname(out_img_location)
    os.makedirs(out_img_dir, exist_ok=True)

    extension = os.path.splitext(out_img_location)[1]
    if extension == ".mhd" or extension == ".mha":
        write_mhd_mha(out_img_location, img_np_x_y, spacing_x_y)
    elif extension == ".png":
        write_png(out_img_location, img_np_x_y)
    else:
        raise Exception(
            "Do not recognize extension {} in location {},\
                     unable to write file".format(
                extension, out_img_location
            )
        )
    if not os.path.isfile(out_img_location):
        raise Exception(
            "Failed to write image to location\
                                         {}".format(
                out_img_location
            )
        )


def write_png(out_img_location, img_np_x_y):
    """
    Write png image to disk.  Only uint16 or uint8 supported
    :param out_img_location: The file location to write to
    :param img_np_x_y: The np array of the image to write
    :return:
    """

    # still using pypng for this because sitk only supports writing uchar and ushort png images
    img_np_y_x = np.transpose(img_np_x_y)
    with open(out_img_location, "wb") as f:
        if img_np_y_x.dtype == "uint16":
            writer = png.Writer(
                width=img_np_y_x.shape[1],
                height=img_np_y_x.shape[0],
                greyscale=True,
                bitdepth=16,
            )
        elif img_np_y_x.dtype == "uint8":
            writer = png.Writer(
                width=img_np_y_x.shape[1],
                height=img_np_y_x.shape[0],
                greyscale=True,
                bitdepth=8,
            )
        else:
            raise Exception(
                "only 8 bit and 16 bit png writing supported.\
                            Please provide np array as uint16 or uint8. \
                            Failed to write {}".format(
                    out_img_location
                )
            )
        img_as_list_rows = img_np_y_x.tolist()
        writer.write(f, img_as_list_rows)


def write_mhd_mha(out_img_location, img_np_x_y, spacing_x_y):
    """
    A method which writes an mhd or mha image
    given the np version in X(axis 0), Y(axis 1) format with spacing
    :param out_img_location: The file location to write to
    :param img_np_x_y: The np array in x, y, width, height format
    :param spacing_x_y: The spacing of the image
    :return:
    """
    itk_out = sitk.GetImageFromArray(np.transpose(img_np_x_y))
    itk_out.SetSpacing(np.transpose(spacing_x_y))

    sitk.WriteImage(itk_out, out_img_location, True)



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 27 15:40:35 2020
@author: keelin
"""

from pathlib import Path
import numpy as np
import opencxr
from opencxr.utils.file_io import read_file


def test_imagesorter():
    # Load the algorithm
    img_sorter_algorithm = opencxr.load(opencxr.algorithms.image_sorter)
    # read in a test image
    # note that the image sorter is designed to run on raw data so it performs better without prior image normalization/standardization
    f_in = Path(__file__).parent / "resources" / "images" / "c0004.mha"
    f_in = str(f_in.resolve())
    img_np, spacing, pydata = read_file(f_in)
    # get the output from the image sorter
    # the output is a dict something like the one shown below.
    # The first four keys give classifications for Type, Rotation, Inversion, Lateral_Flip.
    # The second four keys provide probabilities of all possible classes for users that might need this
    # {'Type': 'PA',
    #  'Rotation': '0',
    #   'Inversion': 'No',
    #   'Lateral_Flip': 'No',
    #   'Type_Probs_PA_AP_lateral_notCXR': [0.99999976, 2.5101654e-08, 2.4382584e-07, 1.0590604e-08],
    #   'Rotation_Probs_0_90_180_270': [0.9999999, 2.7740466e-08, 2.2800064e-08, 3.7591672e-08],
    #   'Inversion_Probs_No_Yes': [0.9999968589511354, 3.1410489e-06],
    #   'Lateral_Flip_Probs_No_Yes': [0.9999986753330177, 1.324667e-06]}

    # Possible values for the first 4 keys listed here are as follows:
    #       Type:  ['PA', 'AP', 'lateral', 'notCXR']
    #       Rotation: ['0', '90', '180', '270']
    #       Inversion: ['No', 'Yes']
    #       Lateral_Flip: ['No', 'Yes']
    result = img_sorter_algorithm.run(img_np)

    expected_result = {
        "Type": "PA",
        "Rotation": "0",
        "Inversion": "No",
        "Lateral_Flip": "No",
        "Type_Probs_PA_AP_lateral_notCXR": [
            0.99999976,
            2.5101654e-08,
            2.4382584e-07,
            1.0590604e-08,
        ],
        "Rotation_Probs_0_90_180_270": [
            0.9999999,
            2.7740466e-08,
            2.2800064e-08,
            3.7591672e-08,
        ],
        "Inversion_Probs_No_Yes": [0.9999968589511354, 3.1410489e-06],
        "Lateral_Flip_Probs_No_Yes": [0.9999986753330177, 1.324667e-06],
    }

    first_four_keys_passed = False
    if (
        result["Type"] == expected_result["Type"]
        and result["Rotation"] == expected_result["Rotation"]
        and result["Inversion"] == expected_result["Inversion"]
        and result["Lateral_Flip"] == expected_result["Lateral_Flip"]
    ):
        first_four_keys_passed = True
    else:
        print("Failed to match first four keys checked", result, expected_result)

    last_four_keys_passed = False

    if (
        np.array_equal(
            np.round(np.asarray(result["Type_Probs_PA_AP_lateral_notCXR"]), 3),
            np.round(np.asarray(expected_result["Type_Probs_PA_AP_lateral_notCXR"]), 3),
        )
        and np.array_equal(
            np.round(np.asarray(result["Rotation_Probs_0_90_180_270"]), 3),
            np.round(np.asarray(expected_result["Rotation_Probs_0_90_180_270"]), 3),
        )
        and np.array_equal(
            np.round(np.asarray(result["Inversion_Probs_No_Yes"]), 3),
            np.round(np.asarray(expected_result["Inversion_Probs_No_Yes"]), 3),
        )
        and np.array_equal(
            np.round(np.asarray(result["Lateral_Flip_Probs_No_Yes"]), 3),
            np.round(np.asarray(expected_result["Lateral_Flip_Probs_No_Yes"]), 3),
        )
    ):
        last_four_keys_passed = True
    else:
        print("Failed to match last four keys checked", result, expected_result)

    passed = first_four_keys_passed and last_four_keys_passed

    if passed:
        print("Image Sorter test completed successfully")
    else:
        print("Image Sorter results not as expected")

    assert passed


if __name__ == "__main__":
    test_imagesorter()



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 11:18:06 2020

@author: keelin
"""

from opencxr.algorithms import create_algorithm


def load(opencxr_algorithm_name):
    algorithm = create_algorithm(opencxr_algorithm_name)
    return algorithm



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 11:19:57 2020

@author: keelin
"""

import numpy as np
from opencxr.utils.mask_crop import crop_img_borders
from opencxr.utils.resize_rescale import (
    rescale_to_min_max,
    resize_long_edge_and_pad_to_square,
)


def clip_at_percentiles(img_np, perc_low, perc_high):
    """
    Clip image values based on intensity percentiles
    :param img_np: input image as np array
    :param perc_low: lower percentile (0-100)
    :param perc_high: upper percentile (0-100)
    :return: the image with clipped values
    """
    in_dtype = img_np.dtype
    low_clip = np.percentile(img_np, perc_low)
    high_clip = np.percentile(img_np, perc_high)
    return np.clip(img_np, low_clip, high_clip).astype(in_dtype)


def preprocess_img(img_x_y):
    """
    Do preprocessing before feeding to imagesorter model
    :param img_x_y: Input image as np array, x,y ordering (eg from utils file_io read_image())
    :return: the preprocessed image, prepared and transposed for the model
    """

    # convert to uint8 with min and max of 0-255
    img_x_y = rescale_to_min_max(img_x_y, np.uint8, 0, 255)

    # crop any homogeneous border regions in the image
    img_x_y, crop_values = crop_img_borders(img_x_y, in_thresh_factor=0.05)

    # clip intensities at 1st and 99th percentiles
    img_x_y = clip_at_percentiles(img_x_y, 1, 99)

    # resize to 256 square image, preserving aspect ratio
    img_x_y, _, _ = resize_long_edge_and_pad_to_square(
        img_x_y, [1, 1], 256, pad_value=0, anti_aliasing=True, interp_order=1
    )

    # re-convert to uint8 since resize returns float
    img_x_y = rescale_to_min_max(img_x_y, new_dtype=np.uint8, new_min=0, new_max=255)

    # transpose before returning since this is what model expects
    return np.transpose(img_x_y)



#!/usr/bin/env python

import os
from typing import Dict

from setuptools import find_packages, setup

NAME = "opencxr"
REQUIRES_PYTHON = ">=3.7"

with open("README.md", "r") as readme_file:
    readme = readme_file.read()


requirements = [
    # "pandas!=0.24.0",
    # "imageio",
    # "cookiecutter",
    # "click",
    # "scipy",
    "wget==3.2",
    "tensorflow==2.11.0",
    "SimpleITK==2.3.0",
    "keras==2.14.0",
    "pydicom==2.4.3",
    "pypng==0.20220715.0",
    "scikit-image==0.22.0",
    "scikit-build==0.17.6",
    "numpy==1.26.0",
]

# test_requirements = ["pytest", "pytest-cov", "pytest-xdist", "pytest-randomly"]

# here = os.path.abspath(os.path.dirname(__file__))

setup(
    name="opencxr",
    author="Keelin Murphy",
    author_email="keelin.murphy@radboudumc.nl",
    description=(
        "a collection of algorithms for processing of chest radiograph (CXR) images"
    ),
    install_requires=requirements,
    # include_package_data=True,
    license="Apache 2.0",
    long_description="a collection of algorithms for processing of chest radiograph (CXR) images",
    keywords="opencxr",
    url="https://github.com/DIAGNijmegen/opencxr",
    packages=find_packages(),
    version="1.3.1",
)



# -*- coding: utf-8 -*-
"""

@author: keelin
"""

from pathlib import Path
import opencxr
from opencxr.utils import apply_size_changes_to_img
from opencxr.utils.file_io import read_file, write_file
import numpy as np


def test_cxrstandardization():
    # read in and standardize an image
    # first get the file path relative to current location
    f_in = Path(__file__).parent / "resources" / "images" / "c0002.mha"
    f_in = str(f_in.resolve())
    # read in the image
    img_np, spacing, pydata = read_file(f_in)
    # load the standardization algorithm
    cxrstandardize_algorithm = opencxr.load(opencxr.algorithms.cxr_standardize)

    # run the standardization algorithm on the test image
    # this will return the new image, the new spacing, and a dict of size changes carried out
    # these size changes can be easily applied to other images, see further code below
    final_norm_img, new_spacing, size_changes = cxrstandardize_algorithm.run(
        img_np, spacing
    )

    # set up an output location and write the output image there
    f_out = Path(__file__).parent / "resources" / "tmp_test_outputs" / "c0002_norm.mha"
    f_out = str(f_out.resolve())
    write_file(f_out, final_norm_img, new_spacing)

    # Next we test whether the size_changes information that was provided is correct
    # run apply_size_changes_to_img and verify that the output sizes are the same as the ones that came from standardization algorithm
    img_resized_to_test, new_spacing_to_test = apply_size_changes_to_img(
        img_np, spacing, size_changes
    )

    resized_img_worked = final_norm_img.shape == img_resized_to_test.shape
    resized_spacing_worked = new_spacing == new_spacing_to_test

    passed = (
        final_norm_img.shape == (1024, 1024)
        and not np.max(final_norm_img) == 0
        and (final_norm_img.shape == img_resized_to_test.shape)
        and (new_spacing == new_spacing_to_test)
    )

    if passed:
        print("CXR Standardization test completed successfully")
    else:
        print("CXR Standardization test failed")

    assert passed


if __name__ == "__main__":
    test_cxrstandardization()



# -*- coding: utf-8 -*-
"""

@author: ecem
"""

from pathlib import Path
import numpy as np

import opencxr
from opencxr.utils.file_io import read_file, write_file


def test_heart_seg():
    # Load the heart segmentation algorithm
    heartseg_algorithm = opencxr.load(opencxr.algorithms.heart_seg)
    # read in a test image
    f_in = Path(__file__).parent / "resources" / "images" / "c0003.mha"
    f_in = str(f_in.resolve())
    img_np, spacing, pydata = read_file(f_in)
    # run the heart segmentation algorithm (note - in the wild it may be best to perform cxr standardization first)
    # this will return the segmentation map image
    seg_map = heartseg_algorithm.run(img_np)
    # write the output from the segmentation algorithm to disk
    f_out = (
        Path(__file__).parent / "resources" / "tmp_test_outputs" / "c0003_heartseg.mha"
    )
    f_out = str(f_out.resolve())
    write_file(f_out, seg_map, spacing)

    passed = seg_map.shape == img_np.shape and not np.max(seg_map) == 0

    if passed:
        print("Heart Segmentation test completed successfully")
    else:
        print("Heart Segmentation results not as expected")

    assert passed


if __name__ == "__main__":
    test_heart_seg()



from opencxr.utils.mask_crop import crop_with_params, uncrop_with_params
from opencxr.utils.resize_rescale import (
    resize_to_x_y,
    un_pad_axis_with_total,
    pad_axis_with_total,
)

size_change_resize_to_x_y = "sc_resize_to_x_y"
size_change_pad_axis_with_total = "sc_pad_axis_with_total"
size_change_unpad_axis_with_total = "sc_unpad_axis_with_total"
size_change_crop_with_params = "sc_crop_with_params"
size_change_uncrop_with_params = "sc_uncrop_with_params"


def apply_size_changes_to_img(
    img_np, spacing, size_changes, anti_aliasing=True, interp_order=1
):
    """
    Applies specified size changes to an image.  Algorithms which change size (e.g cxr standardization) can return size_changes information in list format
    This method is a utility to allow users to apply those same size changes to another image.
    Args:
        img_np: the input image in x,y axis ordering
        size_changes: the size changes returned from an algorithm, to be applied to input image

    Returns:
        The image with size changes applied
        The new spacing for the resized image
    """

    # Iterate over list of operations:
    for i in size_changes:
        # the operation name is the first element, parameters are in second element
        op_to_perform = i[0]
        params = i[1]
        # if the operation is to resize to specified x, y dimension
        if op_to_perform == size_change_resize_to_x_y:
            new_size_0 = params[2]
            new_size_1 = params[3]
            img_np, spacing, _ = resize_to_x_y(
                img_np,
                spacing,
                new_size_0,
                new_size_1,
                anti_aliasing=anti_aliasing,
                interp_order=interp_order,
            )

        # if the operation is to pad a specified axis with constants
        elif op_to_perform == size_change_pad_axis_with_total:
            axis_to_pad = params[0]
            total_pad = params[1]
            pad_value = params[2]
            img_np, _ = pad_axis_with_total(img_np, axis_to_pad, total_pad, pad_value)

        # if the operation is to crop using specified x and y values
        elif op_to_perform == size_change_crop_with_params:
            orig_size_x = params[0]
            orig_size_y = params[1]
            minx = params[2]
            maxx = params[3]
            miny = params[4]
            maxy = params[5]
            img_np, _ = crop_with_params(img_np, [minx, maxx, miny, maxy])

        elif op_to_perform == size_change_unpad_axis_with_total:
            axis = params[0]
            total_pad = params[1]
            img_np, _ = un_pad_axis_with_total(img_np, axis, total_pad)
        else:
            print(
                "ERROR, the operation",
                op_to_perform,
                "is not implemented in apply_size_changes_to_img",
            )
            return None

    return img_np, spacing


def reverse_size_changes_to_img(
    img_np, spacing, size_changes, anti_aliasing=True, interp_order=1
):
    """
    Reverses specified size changes on the given image.  Algorithms which change size (e.g. cxr standardization) can return size_changes information in list format
    This method is a utility to allow users to reverse those size changes at a later stage on any image
    Args:
        img_np: The image to be processed, xy ordering
        size_changes: The list of size changes provided by an algorithm which implemented them

    Returns:
        out_img_np: A resized version of the input image, where the specified size changes have been reversed
        spacing: the new spacing for the resized image

    """
    # Iterate over list in reverse order of operations:
    for i in reversed(size_changes):
        # the operation name is the first element, parameters are in second element
        op_to_perform = i[0]
        params = i[1]
        # if the operation performed was a resize to specific x, y dimensions
        if op_to_perform == size_change_resize_to_x_y:
            # the original size is stored in params 0 and 1
            return_to_size_0 = params[0]
            return_to_size_1 = params[1]
            # return to original size
            img_np, spacing, _ = resize_to_x_y(
                img_np,
                spacing,
                return_to_size_0,
                return_to_size_1,
                anti_aliasing=anti_aliasing,
                interp_order=interp_order,
            )

        # if the operation performed was a padding of a specified axis with constants
        elif op_to_perform == size_change_pad_axis_with_total:
            axis_was_padded = params[0]
            total_pad = params[1]
            # undo the padding operation
            img_np, _ = un_pad_axis_with_total(img_np, axis_was_padded, total_pad)

        # if the operation performed was a crop using specifed x, y parameters
        elif op_to_perform == size_change_crop_with_params:
            # original size is stored in params 0 and 1
            orig_size_x = params[0]
            orig_size_y = params[1]
            minx = params[2]
            maxx = params[3]
            miny = params[4]
            maxy = params[5]
            # undo the crop operation
            img_np, _ = uncrop_with_params(
                img_np, orig_size_x, orig_size_y, [minx, maxx, miny, maxy]
            )

        else:
            print(
                "ERROR, the operation",
                op_to_perform,
                "is not implemented in reverse_size_changes_to_img",
            )
            return None

    return img_np, spacing



# -*- coding: utf-8 -*-
"""
Created on Fri July  2 2021

@author: ecem
"""

from pathlib import Path

import numpy as np
from opencxr.algorithms.base_algorithm import BaseAlgorithm
from opencxr.algorithms.lungsegmentation.model import unet
from opencxr.utils import reverse_size_changes_to_img
from opencxr.utils.mask_crop import tidy_segmentation_mask
from opencxr.utils.resize_rescale import rescale_to_min_max
from opencxr.utils.resize_rescale import resize_long_edge_and_pad_to_square
import wget
import os

"""
Segments the lungs in frontal CXR images
Performance may be better on images that have had cxrstandardization, depending on original image appearance 
"""


class LungSegmentationAlgorithm(BaseAlgorithm):
    def __init__(self):
        """
        initialize the model and weights
        """
        self.model = unet(
            (512, 512, 1),
            k_size=3,
            optimizer="RMSprop",
            depth=6,
            downsize_filters_factor=2,
            batch_norm=True,
            activation="elu",
            initializer="he_normal",
            upsampling=True,
            dropout=False,
            n_convs_per_layer=1,
            lr=0.0007658078359138082,
        )

        path_to_model_file = (
            Path(__file__).parent.parent / "model_weights" / "lung_seg.h5"
        )
        path_to_model_resolved = str(path_to_model_file.resolve())

        file_url = "https://github.com/DIAGNijmegen/opencxr/raw/master/opencxr/algorithms/model_weights/lung_seg.h5"
        # if the file does not exist (it's not included in whl file) then download it from github
        if not os.path.isfile(path_to_model_resolved):
            print("First use of lung segmentation model, downloading the weights......")
            os.makedirs(os.path.dirname(path_to_model_resolved), exist_ok=True)
            wget.download(file_url, path_to_model_resolved)
            if not os.path.isfile(path_to_model_resolved):
                print("Failed to download file from", file_url)
                print(
                    "Please check the URL is valid and the following location is writeable",
                    path_to_model_resolved,
                )
                return
        if (
            not os.stat(path_to_model_resolved).st_size > 150000000
        ):  # model file needs to be larger than 150MB or it was not pulled correctly
            print(
                "Lung Segmentation model file seems incorrect (size ",
                os.stat(path_to_model_resolved).st_size,
                ") downloading the weights......",
                path_to_model_resolved,
            )
            os.remove(
                path_to_model_resolved
            )  # first delete the file, because wget will not overwrite
            wget.download(file_url, path_to_model_resolved)

        self.model.load_weights(path_to_model_resolved)

    def preprocess(self, image):
        """
        Basic mean and std-dev pre processing
        :param image: np array image
        :return: preprocessed np array image
        """
        image = image / (np.max(image) / 2) - 1

        mean_train = -0.00557407
        std_train = 0.51691783
        image -= mean_train
        image /= std_train

        return image

    def process_image(self, input_image, remove_ratio=0.05):
        """
        Calls the model to obtain lung segmentation
        :param input_image: input np array 512x512 y,x ordering
        :return: np array 512x512 of lung segmentation
        """

        input_image = input_image.astype(np.float32)
        input_image = np.expand_dims(input_image, -1)

        if len(input_image.shape) == 3:
            input_image = np.expand_dims(input_image, 0)

        pr_test = self.model.predict(input_image).squeeze()

        pr_test = pr_test > 0.5

        segment_map = np.zeros(pr_test.shape)
        segment_map[pr_test == True] = 255
        segment_map = segment_map.astype(np.uint8)

        return segment_map

    def name(self):
        return "LungSegmentationAlgorithm"

    def resize_to_original(self, seg_map_np, size_changes):
        """
        Resize the segmentation map (512x512) to original image size
        :param seg_map_np: the segmentation map
        :param size_changes: the size changes that were applied to the original image (obtained from utils resize calls)
        :return: the segmentation map with size matching original input
        """
        # Just reverse the size changes that were applied to the original image
        resized_seg_map, _ = reverse_size_changes_to_img(
            seg_map_np, [1, 1], size_changes, anti_aliasing=False, interp_order=0
        )

        resized_seg_map = rescale_to_min_max(resized_seg_map, np.uint8)

        return resized_seg_map

    def run(self, image):
        """
        The call to run lung segmentation on an image
        :param image: The input image as np array, x, y ordering (eg from utils file_io read_file)
        :return: The segmentation image as np array, same size as input image
        """
        # transpose because lung segmentation model requires y, x ordering
        image = np.transpose(image)

        orig_img_shape = image.shape

        image = np.squeeze(image)
        if len(image.shape) > 2 and (image.shape[-1] > 1):
            image = np.mean(image, axis=-1)

        # resize to 512 to fit the model
        resized_img, new_spacing, size_changes = resize_long_edge_and_pad_to_square(
            image, (1, 1), 512
        )
        # do some basic preprocessing
        resized_img = self.preprocess(resized_img)
        # get the segmentation in size 512
        seg_map = self.process_image(resized_img)

        # if the seg map has nothing segmented then future rescaling operations will fail so we return immediately in that case
        if np.max(seg_map) == 0:
            return np.zeros(orig_img_shape).astype(np.uint8)

        # resize the segmentation (512x512) to the same size as original input using record of original size changes
        seg_original = self.resize_to_original(seg_map, size_changes)
        # tidy up by removing holes and keeping two largest connected components
        seg_original = tidy_segmentation_mask(seg_original, nr_components_to_keep=2)
        # transpose to return content the same way it was input
        seg_original = np.transpose(seg_original)

        return seg_original



# -*- coding: utf-8 -*-
"""
Created on Fri July  2 2021

@author: ecem
"""
import os.path
from pathlib import Path

import numpy as np
from opencxr.algorithms.base_algorithm import BaseAlgorithm
from opencxr.algorithms.heartsegmentation.model import unet
from opencxr.utils import reverse_size_changes_to_img
from opencxr.utils.mask_crop import tidy_segmentation_mask
from opencxr.utils.resize_rescale import (
    rescale_to_min_max,
    resize_long_edge_and_pad_to_square,
)
import wget

"""
Segments the heart in frontal CXR images
Performance may be better on images that have had cxrstandardization, depending on original image appearance 
"""


class HeartSegmentationAlgorithm(BaseAlgorithm):
    def __init__(self):
        """
        Initialize the model and weights
        """
        self.model = unet(
            (512, 512, 1),
            k_size=3,
            optimizer="adam",
            depth=6,
            downsize_filters_factor=2,
            batch_norm=True,
            activation="selu",
            initializer="lecun_normal",
            upsampling=False,
            dropout=False,
            n_convs_per_layer=2,
            lr=0.00018521094785555384,
        )

        path_to_model_file = (
            Path(__file__).parent.parent / "model_weights" / "heart_seg.h5"
        )
        path_to_model_resolved = str(path_to_model_file.resolve())

        file_url = "https://github.com/DIAGNijmegen/opencxr/raw/master/opencxr/algorithms/model_weights/heart_seg.h5"
        if os.path.isfile(path_to_model_resolved):
            print("found model file with size", os.stat(path_to_model_resolved).st_size)
        # if the file does not exist (it's not included in whl file) then download it from github
        if not os.path.isfile(path_to_model_resolved):
            print(
                "First use of heart segmentation model, downloading the weights......"
            )
            os.makedirs(os.path.dirname(path_to_model_resolved), exist_ok=True)
            wget.download(file_url, path_to_model_resolved)
            if not os.path.isfile(path_to_model_resolved):
                print("Failed to download file from", file_url)
                print(
                    "Please check the URL is valid and the following location is writeable",
                    path_to_model_resolved,
                )
                return
        if (
            not os.stat(path_to_model_resolved).st_size > 150000000
        ):  # model file needs to be larger than 150MB or it was not pulled correctly
            print(
                "Heart segmentation model file seems incorrect (size ",
                os.stat(path_to_model_resolved).st_size,
                ") downloading the weights......",
                path_to_model_resolved,
            )
            os.remove(
                path_to_model_resolved
            )  # first delete the file, because wget will not overwrite
            wget.download(file_url, path_to_model_resolved)

        self.model.load_weights(path_to_model_resolved)

    def preprocess(self, image):
        """
        Basic mean and std-dev pre processing
        :param image: np array image
        :return: preprocessed np array image
        """
        image = image / (np.max(image) / 2) - 1

        mean_train = -0.00557407
        std_train = 0.51691783
        image -= mean_train
        image /= std_train

        return image

    def process_image(self, input_image):
        """
        Calls the model to obtain heart segmentation
        :param input_image: input np array 512x512 y,x ordering
        :return: np array 512x512 of heart segmentation
        """

        input_image = input_image.astype(np.float32)
        input_image = np.expand_dims(input_image, -1)

        if len(input_image.shape) == 3:
            input_image = np.expand_dims(input_image, 0)

        pr_test = self.model.predict(input_image).squeeze()

        pr_test = pr_test > 0.5

        segment_map = np.zeros(pr_test.shape)
        segment_map[pr_test == True] = 255
        segment_map = segment_map.astype(np.uint8)

        return segment_map

    def name(self):
        return "HeartSegmentationAlgorithm"

    def resize_to_original(self, seg_map_np, size_changes):
        """
        Resize the segmentation map (512x512) to original image size
        :param seg_map_np: the segmentation map
        :param size_changes: the size changes that were applied to the original image (obtained from utils resize calls)
        :return: the segmentation map with size matching original input
        """

        # Just reverse the size changes that were applied to the original image
        resized_seg_map, _ = reverse_size_changes_to_img(
            seg_map_np, [1, 1], size_changes, anti_aliasing=False, interp_order=0
        )
        resized_seg_map = rescale_to_min_max(resized_seg_map, np.uint8)

        return resized_seg_map

    def run(self, image):
        """
        The call to run heart segmentation on an image
        :param image: The input image as np array, x, y ordering (eg from utils file_io read_file)
        :return: The segmentation image as np array, same size as input image
        """
        # transpose because heart segmentation model requires y, x ordering
        image = np.transpose(image)

        orig_img_shape = image.shape

        image = np.squeeze(image)
        if len(image.shape) > 2 and (image.shape[-1] > 1):
            image = np.mean(image, axis=-1)

        # resize to 512 before passing to the model.  Keep record of the size changes for later
        resized_img, new_spacing, size_changes = resize_long_edge_and_pad_to_square(
            image, (1, 1), 512
        )
        # do basic preprocessing
        resized_img = self.preprocess(resized_img)
        # call the model to do the segmentation
        seg_map = self.process_image(resized_img)

        # if the seg map has nothing segmented then later rescaling operations will fail so we return immediately in that case
        if np.max(seg_map) == 0:
            return np.zeros(orig_img_shape).astype(np.uint8)

        # resize the segmentation (size 512) to the same size as original input using record of original size changes
        seg_original = self.resize_to_original(seg_map, size_changes)
        # tidy up by removing holes and keeping largest connected component
        seg_original = tidy_segmentation_mask(seg_original, nr_components_to_keep=1)
        # transpose to return content the same way it was input
        seg_original = np.transpose(seg_original)

        return seg_original



# -*- coding: utf-8 -*-
"""
Created on Fri July  2 2021

@author: keelin
"""

import numpy as np
import opencxr
from opencxr.algorithms.base_algorithm import BaseAlgorithm
from opencxr.utils.mask_crop import crop_to_mask
from opencxr.utils.normalization import Normalizer
from opencxr.utils.resize_rescale import (
    resize_long_edge_and_pad_to_square,
    rescale_to_min_max,
)

"""
CXR Standardization does the following for frontal CXR images:
1) Standardize (Normalize) image intensities using Energy Based Normalization
2) Optionally segment the lungs and crop around the lung box
3) Resize the image to square size specified preserving aspect ratio and padding the shorter edge
"""


class CXRStandardizationAlgorithm(BaseAlgorithm):
    def __init__(self):
        # create an instance of lung segmentation algorithm so we only need to do this once
        self.lung_seg_alg = opencxr.load(opencxr.algorithms.lung_seg)

    def name(self):
        return "CXRStandardizationAlgorithm"

    def run(self, image_np, spacing, do_crop_to_lung_box=True, final_square_size=1024):
        """
        Runs standardization algorithm
        Args:
            image_np: np array, x, y ordering,
            spacing: tuple or list of spacing values
            do_crop_to_lung_box: boolean to indicate whether to crop around the lung segmentation
            final_square_size: The size (pixels) of the final output image (aspect ratio preserved, short side padded)

        Returns:
            standard_image: np array image with standardized intensities, cropped around lungs (if required), resized to final_square_size with padding to preserve aspect ratio
            new_spacing: the new spacing of the image
            size_changes: A dict identifying size changes that took place (to enable users to repeat these on related images if needed, see utils/__init__.py)
        """
        # normalize intensities (this will also crop black borders and resize image to width 2048 and aspect preserved height)
        (
            norm_img_np,
            new_spacing,
            size_changes_in_norm,
        ) = Normalizer.do_full_normalization(image_np, spacing, self.lung_seg_alg)

        if (
            np.max(norm_img_np) == 0
        ):  # if we didn't get a valid lung segmentation then an empty image comes back:
            # then we won't do cropping to lung box - it will cause errors and makes  no sense anyway!
            do_crop_to_lung_box = False

        if (
            not do_crop_to_lung_box
        ):  # no cropping to lungs, just resize to specified size and exit
            # resize it to square specified size
            (
                norm_np_resized,
                newest_spacing,
                size_changes_square_pad,
            ) = resize_long_edge_and_pad_to_square(
                norm_img_np, new_spacing, final_square_size
            )
            norm_np_resized = np.clip(norm_np_resized, 0, 4095).astype(np.uint16)
            # and join all the size changes that were made, in order
            size_changes_final = size_changes_in_norm + size_changes_square_pad
            return norm_np_resized, newest_spacing, size_changes_final

        elif (
            do_crop_to_lung_box
        ):  # if we found a valid std image and we are supposed to crop to lungs
            # segment lungs on normalized image
            lung_mask_np = self.lung_seg_alg.run(norm_img_np)
            if (
                not np.max(lung_mask_np) == 0
            ):  # if all is good and we found a lung segmentation
                # crop to lung borders
                # first need to make the lung mask binary
                lung_mask_np = rescale_to_min_max(lung_mask_np, new_min=0, new_max=1)
                # then crop to the mask with margin
                lung_cropped_np, size_changes_lung_crop = crop_to_mask(
                    norm_img_np, new_spacing, lung_mask_np, margin_in_mm=15.0
                )

                # now resize the normalized image to max dim specified, preserving aspect ratio and padding to square img.
                (
                    final_lung_cropped_resized,
                    newest_spacing,
                    size_changes_square_pad,
                ) = resize_long_edge_and_pad_to_square(
                    lung_cropped_np, new_spacing, final_square_size
                )
                final_lung_cropped_resized = np.clip(
                    final_lung_cropped_resized, 0, 4095
                ).astype(np.uint16)
                # and join all the size changes that were made, in order
                size_changes_final = (
                    size_changes_in_norm
                    + size_changes_lung_crop
                    + size_changes_square_pad
                )
                return final_lung_cropped_resized, newest_spacing, size_changes_final

            else:  # no valid lung segmentation found, return an empty image
                # resize it to square specified size
                norm_img_np.fill(0)
                (
                    norm_np_resized,
                    newest_spacing,
                    size_changes_square_pad,
                ) = resize_long_edge_and_pad_to_square(
                    norm_img_np, new_spacing, final_square_size
                )
                norm_np_resized = np.clip(norm_np_resized, 0, 4095).astype(np.uint16)
                # and join all the size changes that were made, in order
                size_changes_final = size_changes_in_norm + size_changes_square_pad
                return norm_np_resized, newest_spacing, size_changes_final



# -*- coding: utf-8 -*-
"""
Created on Fri July  2 2021

@author: ecem
"""

import keras
import numpy as np
import tensorflow as tf
from keras.layers import (
    Input,
    Conv2D,
    MaxPooling2D,
    Conv2DTranspose,
    Concatenate,
    Dropout,
    UpSampling2D,
    BatchNormalization,
)
from keras.models import Model

"""=================================================================================================
Model architecture for heart segmentation
With the dropout settings you can specify the depth (-1 is the deepest layer) and the probability.
==================================================================================================="""


def unet(
    x_in,
    k_size=3,
    optimizer="adam",
    depth=4,
    downsize_filters_factor=1,
    batch_norm=False,
    activation="relu",
    initializer="glorot_uniform",
    seed=42,
    upsampling=False,
    dropout=[(-1, 0.5)],
    n_convs_per_layer=2,
    lr=False,
    loss="binary",
):
    # Fix the seed into given argument.
    np.random.seed(seed)

    # if True, deepest layer with 0.5 probability.
    if dropout == True:
        dropout = [(-1, 0.5)]

    # Define the hyperparameter settings.
    # Max number of init filters: 64
    settings = {
        "n_classes": 2,  # nb of classes always 2.
        "depth": depth,  # unet depth 4
        "filters": 64 / downsize_filters_factor,  # unet filters 64
        "kernel_size": (k_size, k_size),  # unet kernel size (3,3)
        "pool_size": (2, 2),  # standard 2,2
        "n_convs_per_layer": n_convs_per_layer,  # standard perlayer 2 conv.
        "activation": activation,  # relu standard
        "kernel_initializer": initializer,  # he_normal standard
        "padding": "same",  # valid standard
        "dropout": dropout,  # dropout is the deepest layer.
        "batch_norm": batch_norm,  # if to use batch norm or not.
        "upsampling": upsampling,  # if true upsampling, else conv2dtranspose.
    }

    # K.clear_session()
    # x_in: input layer, data format: (batch, height, width, channels)
    data = Input(shape=x_in)  # input layer
    layers = {}  # hold all the layers in a dictionary.
    l = data  # input dimension.

    def conv(filters):
        """
        Return convolutional layer given setting arguments.
        """
        return Conv2D(
            filters=filters,
            kernel_size=settings["kernel_size"],
            activation=settings["activation"],
            kernel_initializer=settings["kernel_initializer"],
            padding=settings["padding"],
        )

    def dropout(rate):
        """
        Return dropout layer given rate.
        """
        return Dropout(rate)

    def batchnorm():
        """
        Return batchnormalization layer.
        """
        return BatchNormalization()

    def pool():
        """
        Return pooling layer given setting argument.
        """
        return MaxPooling2D(pool_size=settings["pool_size"])

    def concat():
        """
        Return Concatenate function.
        """
        return Concatenate()

    # transpose convolutional filter.
    def t_conv(filters):

        if upsampling:
            # print('upsampling through neareset neighbor')
            return UpSampling2D(size=settings["pool_size"])

        else:
            # print('upsampling with conv2d learning transpose filter')
            return Conv2DTranspose(
                filters=filters,
                kernel_size=settings["pool_size"],
                strides=settings["pool_size"],
                kernel_initializer=settings["kernel_initializer"],
                padding=settings["padding"],
            )

    # add a new layer to the existing architecture.
    def add(layer, l_in, name):
        """
        Add a new layer to the dictionary given name.
        """
        layers[name] = layer(l_in)
        return layers[name]

    depths = list(range(settings["depth"]))  # [0,1,2,3]

    if settings["dropout"] != False:
        dropout_depths = list(range(settings["depth"] + 1))  # [1,2,3,4]
        dropouts = {
            dropout_depths[d]: rate for d, rate in settings["dropout"]
        }  # apply to the last layer with 0.5 rate..

    contracting_outputs = {}
    # contracting.
    for i in depths:
        for j in range(settings["n_convs_per_layer"]):
            # number of filters.
            n = int(settings["filters"] * (2 ** i))
            # create the layer given previous , add it to dictionary of layers.
            l = add(conv(n), l, "conv_down_{}_{}".format(i, j))

            contracting_outputs[i] = l

            if settings["batch_norm"] == True:
                l = add(batchnorm(), l, "batch_norm_{}_{}".format(i, j))
                contracting_outputs[i] = l

            if settings["dropout"] != False:
                if i in dropouts:
                    l = add(dropout(dropouts[i]), l, "dropout_{}_{}".format(i, j))
                    contracting_outputs[i] = l

        l = add(pool(), l, "pool_{}".format(i))

    # bottom
    i = settings["depth"]
    for j in range(settings["n_convs_per_layer"]):
        n = int(settings["filters"] * 2 ** settings["depth"])
        l = add(conv(n), l, "conv_{}_{}".format(settings["depth"], j))
        # print('layer shape', l.shape)

        if settings["dropout"] != False:
            if i in dropouts:
                l = add(dropout(dropouts[i]), l, "dropout_{}_{}".format(i, j))

    # expanding
    for i in reversed(depths):
        n = int(settings["filters"] * 2 ** i)
        l = add(t_conv(n), l, "t_conv{}".format(i))
        l = add(concat(), [l, contracting_outputs[i]], "concat_{}".format(i))
        for j in range(settings["n_convs_per_layer"]):
            n = int(settings["filters"] * 2 ** i)
            l = add(conv(n), l, "conv_up_{}_{}".format(i, j))

            if settings["batch_norm"] == True:
                l = add(batchnorm(), l, "batch_norm_{}_{}".format(i, j))
                contracting_outputs[i] = l

    lr = float(lr)
    if lr != "default":

        if optimizer == "adam":
            optimizer = tf.keras.optimizers.Adam(learning_rate=lr)
        elif optimizer == "RMSprop":
            optimizer = keras.optimizers.legacy.RMSprop(learning_rate=lr, decay=1e-3)
        elif optimizer == "adagrad":
            optimizer = keras.optimizers.Adagrad(learning_rate=lr)
        elif optimizer == "sgd":
            optimizer = keras.optimizers.legacy.SGD(
                learning_rate=lr, decay=1e-6, momentum=0.9, nesterov=True
            )

    else:  # use default values for the optimizers.
        if optimizer == "sgd":
            optimizer = keras.optimizers.legacy.SGD(
                learning_rate=0.01, decay=1e-6, momentum=0.9, nesterov=True
            )

    if loss == "binary":
        out = Conv2D(filters=1, kernel_size=(1, 1), activation="sigmoid")
        l = add(out, l, "out")
        model = Model(data, l)
        model.compile(
            optimizer=optimizer, loss="binary_crossentropy", metrics=["accuracy"]
        )

    else:  # if not binary classification
        out = Conv2D(filters=2, kernel_size=(1, 1), activation="softmax")
        l = add(out, l, "out")
        model = Model(data, l)
        model.compile(
            optimizer=optimizer, loss="categorical_crossentropy", metrics=["accuracy"]
        )

    return model  # l, layers
